/* ------------------------------------------------------------------------ */
/* hmdgl.c - Funktionen zum Loesen von Differentialgleichungen              */
/*                                                                          */
/*                  Copyright Tilman Kuepper 2017.                          */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 12. Juni 2017                                        */
/* ------------------------------------------------------------------------ */

#include "hmdgl.h"
// #include <malloc.h>  12.06.2017: malloc/free sind in stdlib.h deklariert.
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktion: Array mit <dim> double-Elementen anlegen.         */
/* ------------------------------------------------------------------------ */
static double* dgl_array_malloc(size_t dim)
{
    double *ptr;
    assert(dim > 0);
    ptr = (double*)malloc(dim * sizeof(double));
    if(!ptr) { assert(!"out of memory"); abort(); }
    return ptr;
}

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktion: ein oder mehrere Arrays wieder freigeben.         */
/* ------------------------------------------------------------------------ */
void dgl_array_free(double *parray, ...)
{
  va_list list;
  double *p;
 
  va_start(list, parray); 
  assert(parray);
  for (p = parray; p != 0; p = va_arg(list, double*)) free(p);
  va_end(list);
}

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktion: res = factor_a * a + factor_b * b berechnen.      */
/* ------------------------------------------------------------------------ */
static void dgl_array_add(double res[], size_t dim,
    double a[], double factor_a, double b[], double factor_b)
{
    size_t i;
    assert(res); assert(a); assert(b); assert(dim);
    for(i = 0; i < dim; ++i) res[i] = factor_a * a[i] + factor_b * b[i];
}

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem einfachen            */
/* Euler-Verfahren numerisch integriert. Die Startwerte werden durch die    */
/* Callback-Funktion <init> gesetzt, nach jedem Integrationsschritt wird    */
/* die Callback-Funktion <step> aufgerufen. Ueber den Parameter pdata       */
/* koennen beliebige - vom Anwender definierte - Daten weitergeleitet       */
/* werden (falls nicht erforderlich: pdata = 0 setzen).                     */
/* ------------------------------------------------------------------------ */
void dgl_euler(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata)
{
    size_t i;
    double h  = (t_end - t_begin) / num_steps, t = t_begin;
    double *u = dgl_array_malloc(dim), *k1 = dgl_array_malloc(dim);

    assert(t_end > t_begin && num_steps > 0);
    assert(dim); assert(sys); assert(step); assert(init);

    init(u, pdata);
    for(i = 0; i < num_steps; ++i)
    {
        sys(k1, t, u, pdata);
        dgl_array_add(u, dim, u, 1, k1, h);
        step(t += h, u, pdata);
    }
    dgl_array_free(u, k1, 0);
}

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem Heun-Verfahren       */
/* numerisch integriert. Die Startwerte werden durch die Callback-Funktion  */
/* <init> gesetzt, nach jedem einzelnen Integrationsschritt wird die        */
/* Callback-Funktion <step> aufgerufen. Ueber den Parameter pdata koennen   */
/* beliebige - vom Anwender definierte - Daten weitergeleitet werden        */
/* (falls nicht erforderlich: pdata = 0 setzen).                            */
/* ------------------------------------------------------------------------ */
void dgl_heun(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata)
{
    size_t i;
    double h   = (t_end - t_begin) / num_steps, t = t_begin;
    double *u  = dgl_array_malloc(dim), *tmp = dgl_array_malloc(dim);
    double *k1 = dgl_array_malloc(dim), *k2  = dgl_array_malloc(dim);

    assert(t_end > t_begin && num_steps > 0);
    assert(sys); assert(step); assert(init); assert(dim);

    /*  Siehe de.wikipedia.org: Heun-Verfahren */
    init(u, pdata);
    for(i = 0; i < num_steps; ++i)
    {
        sys(k1, t, u, pdata);
        dgl_array_add(tmp, dim, u, 1, k1, h);
        sys(k2, t + h, tmp, pdata);

        dgl_array_add(u, dim, u, 1, k1, h / 2);
        dgl_array_add(u, dim, u, 1, k2, h / 2);
        step(t += h, u, pdata);
    }
    dgl_array_free(u, tmp, k1, k2, 0);
}

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem klassischen          */
/* 4-stufigen Runge-Kutta-Verfahren numerisch integriert. Die Startwerte    */
/* werden durch die Callback-Funktion <init> gesetzt, nach jedem            */
/* Integrationsschritt wird die Callback-Funktion <step> aufgerufen. Ueber  */
/* den Parameter pdata koennen beliebige - vom Anwender definierte - Daten  */
/* weitergeleitet werden (falls nicht erforderlich: pdata = 0 setzen).      */
/* ------------------------------------------------------------------------ */
void dgl_rk4(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata)
{
    /* Diese Implementierung basiert auf:
     * Arens et al.: Mathematik, 1. korrigierter Nachdruck, S. 433f. */
    size_t i;
    double h   = (t_end - t_begin) / num_steps, t = t_begin;
    double *u  = dgl_array_malloc(dim), *tmp = dgl_array_malloc(dim);
    double *k1 = dgl_array_malloc(dim), *k2  = dgl_array_malloc(dim);
    double *k3 = dgl_array_malloc(dim), *k4  = dgl_array_malloc(dim);

    assert(t_end > t_begin && num_steps > 0);
    assert(sys); assert(step); assert(init); assert(dim);

    init(u, pdata);
    for(i = 0; i < num_steps; ++i)
    {
        sys(k1, t, u, pdata);
        dgl_array_add(tmp, dim, u, 1, k1, h/2); sys(k2, t + h/2, tmp, pdata);
        dgl_array_add(tmp, dim, u, 1, k2, h/2); sys(k3, t + h/2, tmp, pdata);
        dgl_array_add(tmp, dim, u, 1, k3, h/1); sys(k4, t + h/1, tmp, pdata);

        dgl_array_add(u, dim, u, 1, k1, 1*h/6);
        dgl_array_add(u, dim, u, 1, k2, 2*h/6);
        dgl_array_add(u, dim, u, 1, k3, 2*h/6);
        dgl_array_add(u, dim, u, 1, k4, 1*h/6);
        step(t += h, u, pdata);
    }
    dgl_array_free(u, tmp, k1, k2, k3, k4, 0);
}
