/* ------------------------------------------------------------------------ */
/* hmdgl.h - Funktionen zum Loesen von Differentialgleichungen              */
/*                                                                          */
/*                  Copyright Tilman Kuepper 2017.                          */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 12. Juni 2017                                        */
/* ------------------------------------------------------------------------ */

#ifndef HMDGL_Included
#define HMDGL_Included

#include <stdlib.h>  /* size_t */

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------ */
/* Definition des Differentialgleichungssystems: Eine Callback-Funktion     */
/* des Typs <dgl_sys_fnc> berechnet die rechte Seite von du/dt = f(t, u)    */
/* und speichert das Ergebnis im Vektor dudt[]. Im Parameter t wird die     */
/* aktuelle Simulationszeit uebergeben, ueber den Parameter pdata koennen   */
/* beliebige - vom Anwender definierte - Daten weitergeleitet werden.       */
/* ------------------------------------------------------------------------ */
typedef void (*dgl_sys_fnc)(
    double dudt[], double t, const double u[], void *pdata);

/* ------------------------------------------------------------------------ */
/* Nach jedem einzelnen Integrationsschritt wird eine Callback-Funktion     */
/* des Typs <dgl_step_fnc> aufgerufen und die aktuellen Werte von t und     */
/* u[] uebergeben. Ueber den Parameter pdata koennen beliebige - vom        */
/* Anwender definierte - Daten weitergeleitet werden.                       */
/* ------------------------------------------------------------------------ */
typedef void (*dgl_step_fnc)(double t, const double u[], void *pdata);

/* ------------------------------------------------------------------------ */
/* Vor Beginn des eigentlichen Loesungsverfahrens werden die Startwerte     */
/* der verwendeten Variablen durch eine Callback-Funktion des Typs          */
/* <dgl_init_fnc> gesetzt. Ueber den Parameter pdata koennen beliebige -    */
/* vom Anwender definierte - Daten weitergeleitet werden.                   */
/* ------------------------------------------------------------------------ */
typedef void (*dgl_init_fnc)(double u[], void *pdata);

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem einfachen            */
/* Euler-Verfahren numerisch integriert. Die Startwerte werden durch die    */
/* Callback-Funktion <init> gesetzt, nach jedem Integrationsschritt wird    */
/* die Callback-Funktion <step> aufgerufen. Ueber den Parameter pdata       */
/* koennen beliebige - vom Anwender definierte - Daten weitergeleitet       */
/* werden (falls nicht erforderlich: pdata = 0 setzen).                     */
/* ------------------------------------------------------------------------ */
void dgl_euler(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata);

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem Heun-Verfahren       */
/* numerisch integriert. Die Startwerte werden durch die Callback-Funktion  */
/* <init> gesetzt, nach jedem einzelnen Integrationsschritt wird die        */
/* Callback-Funktion <step> aufgerufen. Ueber den Parameter pdata koennen   */
/* beliebige - vom Anwender definierte - Daten weitergeleitet werden        */
/* (falls nicht erforderlich: pdata = 0 setzen).                            */
/* ------------------------------------------------------------------------ */
void dgl_heun(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata);

/* ------------------------------------------------------------------------ */
/* Das Differentialgleichungssystem <sys> wird mit dem klassischen          */
/* 4-stufigen Runge-Kutta-Verfahren numerisch integriert. Die Startwerte    */
/* werden durch die Callback-Funktion <init> gesetzt, nach jedem            */
/* Integrationsschritt wird die Callback-Funktion <step> aufgerufen. Ueber  */
/* den Parameter pdata koennen beliebige - vom Anwender definierte - Daten  */
/* weitergeleitet werden (falls nicht erforderlich: pdata = 0 setzen).      */
/* ------------------------------------------------------------------------ */
void dgl_rk4(double t_begin, double t_end, size_t num_steps, size_t dim,
    dgl_sys_fnc sys, dgl_step_fnc step, dgl_init_fnc init, void *pdata);

#ifdef __cplusplus
}
#endif

#endif  /* HMDGL_Included */
