// ---------------------------------------------------------------------------
// hmdgl.hpp - C++-Funktionen zum Loesen von Differentialgleichungen
//
//                  Copyright Tilman Kuepper 2017.
//    Distributed under the Boost Software License, Version 1.0.
//        (See accompanying file LICENSE_1_0.txt or copy at
//               http://www.boost.org/LICENSE_1_0.txt)
//
// Letzte Bearbeitung: 12. Juni 2017
// ---------------------------------------------------------------------------

#ifndef HMDGL_HPP_Included
#define HMDGL_HPP_Included

// Hinweis:
// Die Zustandsvariablen des Differentialgleichungssystems werden in
// Containern des Typs <state_type> gespeichert. Als <state_type> sind
// grundsaetzlich alle Container geeignet, die die Methoden begin()
// und end() besitzen, zum Beispiel std::vector<double>.

#include <stdlib.h>  // size_t
#include <assert.h>

// Namespace fuer HMDGL-Funktionen
namespace HMDGL
{
    // Namespace fuer interne Hilfsfunktionen
    namespace Detail
    {
        // Hilfsfunktion zur Berechnung von (a_vect + factor_b * b_vect)
        template<class state_type> state_type add(
            state_type a_vect, double factor_b, const state_type& b_vect)
        {
            auto a_it = a_vect.begin(), a_end = a_vect.end();
            auto b_it = b_vect.begin();

            for(; a_it != a_end; ++a_it, ++b_it)
                (*a_it) += factor_b * (*b_it);
            return a_vect;
        }
    }  // namespace Detail

    // -----------------------------------------------------------------------
    // Das Differentialgleichungssystem <sys> wird mit dem Euler-Verfahren
    // numerisch geloest. Die Anfangswerte werden in <init> uebergeben,
    // nach jedem Integrationsschritt wird die Funktion <step> aufgerufen.
    // -----------------------------------------------------------------------
    template<class sys_fn, class step_fn, class state_type>
    void dgl_euler(double t_begin, double t_end, size_t num_steps,
        sys_fn sys, step_fn step, state_type init)
    {
        assert(t_end > t_begin && num_steps > 0);
        double h = (t_end - t_begin) / num_steps, t = t_begin;
        using Detail::add;

        state_type u = init, k1 = init;

        // Siehe de.wikipedia.org: Explizites Euler-Verfahren
        for(size_t i = 0; i < num_steps; ++i)
        {
            k1 = sys(t, u);
            u  = add(u, h, k1);
            step(t += h, u);
        }
    }

    // -----------------------------------------------------------------------
    // Das Differentialgleichungssystem <sys> wird mit dem Heun-Verfahren
    // numerisch geloest. Die Anfangswerte werden in <init> uebergeben,
    // nach jedem Integrationsschritt wird die Funktion <step> aufgerufen.
    // -----------------------------------------------------------------------
    template<class sys_fn, class step_fn, class state_type>
    void dgl_heun(double t_begin, double t_end, size_t num_steps,
        sys_fn sys, step_fn step, state_type init)
    {
        assert(t_end > t_begin && num_steps > 0);
        double h = (t_end - t_begin) / num_steps, t = t_begin;
        using Detail::add;

        state_type u  = init, tmp = init;
        state_type k1 = init, k2  = init;

        // Siehe de.wikipedia.org: Heun-Verfahren
        for(size_t i = 0; i < num_steps; ++i)
        {
            k1  = sys(t, u);
            tmp = add(u, h, k1);
            k2  = sys(t + h, tmp);

            u = add(u, h / 2, k1);
            u = add(u, h / 2, k2);
            step(t += h, u);
        }
    }

    // -----------------------------------------------------------------------
    // Das Differentialgleichungssystem <sys> wird mit dem klassischen
    // vierstufigen Runge-Kutta-Verfahren numerisch geloest. Die Anfangswerte
    // werden in <init> uebergeben, nach jedem Integrationsschritt wird die
    // Funktion <step> aufgerufen.
    // -----------------------------------------------------------------------
    template<class sys_fn, class step_fn, class state_type>
    void dgl_rk4(double t_begin, double t_end, size_t num_steps,
        sys_fn sys, step_fn step, state_type init)
    {
        assert(t_end > t_begin && num_steps > 0);
        double h = (t_end - t_begin) / num_steps, t = t_begin;
        using Detail::add;

        state_type u  = init, tmp = init;
        state_type k1 = init, k2  = init;
        state_type k3 = init, k4  = init;

        // Diese Implementierung basiert auf:
        // Arens et al.: Mathematik, 1. korrigierter Nachdruck, S. 433f.
        for(size_t i = 0; i < num_steps; ++i)
        {
            k1 = sys(t, u);
            tmp = add(u, h / 2, k1); k2 = sys(t + h / 2, tmp);
            tmp = add(u, h / 2, k2); k3 = sys(t + h / 2, tmp);
            tmp = add(u, h / 1, k3); k4 = sys(t + h / 1, tmp);

            u = add(u, 1 * h / 6, k1);
            u = add(u, 2 * h / 6, k2);
            u = add(u, 2 * h / 6, k3);
            u = add(u, 1 * h / 6, k4);
            step(t += h, u);
        }
    }
}  // namespace HMDGL

#endif  // HMDGL_HPP_Included
